import argparse
from typing import List

from wai.logging import LOGGING_WARNING

from seppl import Plugin
from idc.api import ImageData, StreamWriter, DATATYPES, data_type_to_class, BatchWriter, make_list


class MultiWriter(StreamWriter):

    def __init__(self, writers: List[str] = None, data_type: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param writers: the base writers to use (command-line)
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param data_type: the type of output to accept
        :type data_type: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.writers = writers
        self.data_type = data_type
        self._writers = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-multi"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Forwards the incoming data to all the base writers."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-w", "--writer", type=str, default=None, help="The command-line defining the base writer.", required=True, nargs="+")
        parser.add_argument("-t", "--data_type", choices=DATATYPES, type=str, default=None, help="The type of data to accept", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.writers = ns.writer
        self.data_type = ns.data_type

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        if self.data_type is None:
            return [ImageData]
        else:
            return [data_type_to_class(self.data_type)]

    def _parse_commandline(self, cmdline: str) -> List[Plugin]:
        """
        Parses the command-line and returns the list of plugins it represents.
        Raises an exception in case of an invalid sub-flow.

        :param cmdline: the command-line to parse
        :type cmdline: str
        :return:
        """
        from idc.registry import available_writers
        from seppl import args_to_objects, split_args, split_cmdline

        # split command-line into valid plugin subsets
        valid = available_writers()
        args = split_args(split_cmdline(cmdline), list(valid.keys()))
        return args_to_objects(args, valid, allow_global_options=False)

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.data_type is None:
            raise Exception("No data type defined!")
        if (self.writers is None) or (len(self.writers) == 0):
            raise Exception("No writer(s) defined!")
        self._writers = []
        for writer in self.writers:
            objs = self._parse_commandline(writer)
            if len(objs) == 1:
                _writer = objs[0]
                if self.accepts()[0] in _writer.accepts():
                    self._writers.append(_writer)
                else:
                    raise Exception("Writer '%s' accepts '%s' but '%s' is required!" % (writer, str(_writer.accepts()), str(self.accepts())))
            else:
                raise Exception("Failed to obtain a single writer from command-line: %s" % writer)
        for writer in self._writers:
            writer.initialize()
            writer.session = self.session
        self.logger().info("# writers: %d" % len(self._writers))

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for writer in self._writers:
            if isinstance(writer, StreamWriter):
                writer.write_stream(data)
            elif isinstance(writer, BatchWriter):
                writer.write_batch(make_list(data))
            else:
                raise Exception("Unknown type of writer: %s" % str(type(writer)))

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        if self._writers is not None:
            for writer in self._writers:
                writer.finalize()



import argparse
from typing import List, Iterable, Union

from wai.logging import LOGGING_WARNING
from wai.common.geometry import Point, Polygon
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from opex import ObjectPredictions
from seppl.placeholders import PlaceholderSupporter, placeholder_list
from seppl.io import locate_files
from idc.api import ObjectDetectionData, locate_image
from idc.api import Reader


class OPEXObjectDetectionReader(Reader, PlaceholderSupporter):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 resume_from: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param resume_from: the file to resume from (glob)
        :type resume_from: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.resume_from = resume_from
        self._inputs = None
        self._current_input = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-opex-od"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the bounding box and/or polygon definitions from the associated .json file in OPEX format."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the JSON file(s) to read; glob syntax is supported; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the JSON files to use; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("--resume_from", type=str, help="Glob expression matching the file to resume from, e.g., '*/012345.json'", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.resume_from = ns.resume_from

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.json", resume_from=self.resume_from)

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input
        self.logger().info("Reading from: " + str(self.session.current_input))
        img = locate_image(self.session.current_input)
        if img is None:
            self.logger().warning("No corresponding image found for: %s" % self.session.current_input)
            yield None

        preds = ObjectPredictions.load_json_from_file(self.session.current_input)

        lobjs = LocatedObjects()
        for obj in preds.objects:
            meta = dict()
            if obj.score is not None:
                meta["score"] = obj.score
            if obj.label is not None:
                meta["type"] = obj.label
            if isinstance(obj.meta, dict):
                for k, v in obj.meta.items():
                    meta[k] = v
            lobj = LocatedObject(obj.bbox.left, obj.bbox.top, obj.bbox.right - obj.bbox.left + 1, obj.bbox.bottom - obj.bbox.top + 1, **meta)
            lpoints = []
            for point in obj.polygon.points:
                lpoint = Point(point[0], point[1])
                lpoints.append(lpoint)
            lobj.set_polygon(Polygon(*lpoints))
            lobjs.append(lobj)

        # meta-data?
        meta = dict()
        meta["id"] = preds.id
        if preds.timestamp is not None:
            meta["timestamp"] = preds.timestamp
        if isinstance(preds.meta, dict):
            meta = dict()
            for k, v in preds.meta.items():
                meta[k] = v

        yield ObjectDetectionData(source=img, annotation=lobjs, metadata=meta)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0



import argparse
import os.path
from typing import List

from seppl.io import Filter
from simple_palette_utils import generate_palette_list, PALETTE_AUTO, palettes
from wai.logging import LOGGING_WARNING

from idc.api import DATATYPES, data_type_to_class, ImageData, to_indexedpng
from idc.api import ImageSegmentationData, flatten_list, make_list


class UseMask(Filter):
    """
    Uses the images segmentation annotations (= mask) as the new base image.
    """

    def __init__(self, data_type: str = None, palette: str = None, use_rgb: bool = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param data_type: the type of output to generate with the mask
        :type data_type: str
        :param palette: the palette to use, either a supported palette name (auto|x11|light|dark) or comma-separated list of R,G,B values
        :type palette: str
        :param use_rgb: whether to use RGB mode rather than palette mode
        :type use_rgb: bool
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.data_type = data_type
        self.palette = palette
        self.use_rgb = use_rgb
        self._output_cls = None
        self._palette_list = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "use-mask"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Uses the images segmentation annotations (= mask) as the new base image."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageSegmentationData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        if self.data_type is None:
            return [ImageData]
        else:
            return [data_type_to_class(self.data_type)]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-t", "--data_type", choices=DATATYPES, type=str, default=None, help="The type of data to forward", required=True)
        parser.add_argument("-p", "--palette", metavar="PALETTE", type=str, default=PALETTE_AUTO, help="The palette to use; either palette name (%s) or comma-separated list of R,G,B values." % "|".join(palettes()), required=False)
        parser.add_argument("--use_rgb", action="store_true", help="Whether to force RGB mode instead of palette mode.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.data_type = ns.data_type
        self.palette = ns.palette
        self.use_rgb = ns.use_rgb

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.data_type is None:
            raise Exception("No data type defined!")
        self._palette_list = generate_palette_list(self.palette)
        self._output_cls = data_type_to_class(self.data_type)

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            if item.has_annotation():
                new_image_name = os.path.splitext(item.image_name)[0] + ".png"
                new_image = to_indexedpng(item.image_width, item.image_height, item.annotation, self._palette_list, background=0)
                if self.use_rgb:
                    new_image = new_image.convert("RGB")
                new_item = self._output_cls(image_name=new_image_name, image=new_image)
                result.append(new_item)

        return flatten_list(result)



import argparse
from typing import List

from wai.logging import LOGGING_WARNING
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from seppl.io import Filter
from idc.api import ObjectDetectionData, flatten_list, make_list


class DimensionDiscarder(Filter):
    """
    Removes annotations which fall outside certain dimensional limits.
    """

    def __init__(self, min_width: int = None, min_height: int = None, 
                 max_width: int = None, max_height: int = None,
                 min_area: int = None, max_area: int = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param min_width: the minimum width for annotations, ignored if None
        :type min_width: int
        :param min_height: the minimum height for annotations, ignored if None
        :type min_height: int
        :param max_width: the maximum width for annotations, ignored if None
        :type max_width: int
        :param max_height: the maximum height for annotations, ignored if None
        :type max_height: int
        :param min_area: the minimum area (polygon if available or bbox) for annotations, ignored if None
        :type min_area: int
        :param max_area: the maximum area (polygon if available or bbox) for annotations, ignored if None
        :type max_area: int
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.min_width = min_width
        self.min_height = min_height
        self.max_width = max_width
        self.max_height = max_height
        self.min_area = min_area
        self.max_area = max_area

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "dimension-discarder"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Removes annotations which fall outside certain dimensional limits."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("--min_width", type=int, default=None, help="The minimum width for annotations", required=False)
        parser.add_argument("--min_height", type=int, default=None, help="The minimum height for annotations", required=False)
        parser.add_argument("--max_width", type=int, default=None, help="The maximum width for annotations", required=False)
        parser.add_argument("--max_height", type=int, default=None, help="The maximum height for annotations", required=False)
        parser.add_argument("--min_area", type=int, default=None, help="The minimum area for annotations (polygon if available, otherwise bbox)", required=False)
        parser.add_argument("--max_area", type=int, default=None, help="The maximum area for annotations (polygon if available, otherwise bbox)", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.min_width = ns.min_width
        self.min_height = ns.min_height
        self.max_width = ns.max_width
        self.max_height = ns.max_height
        self.min_area = ns.min_area
        self.max_area = ns.max_area

    def _should_discard_located_object(self, located_object: LocatedObject) -> bool:
        """
        Decides if the located object should be discarded.

        :param located_object:  The located object.
        :return:                True if it should be discarded,
                                False if it should be kept.
        """
        # Min width check
        if self.min_width is not None and located_object.width < self.min_width:
            self.logger().debug("Width too small: %d < %d" % (located_object.width, self.min_width))
            return True

        # Max width check
        if self.max_width is not None and located_object.width > self.max_width:
            self.logger().debug("Width too large: %d > %d" % (located_object.width, self.max_width))
            return True

        # Min height check
        if self.min_height is not None and located_object.height < self.min_height:
            self.logger().debug("Height too small: %d < %d" % (located_object.height, self.min_height))
            return True

        # Max height check
        if self.max_height is not None and located_object.height > self.max_height:
            self.logger().debug("Height too large: %d > %d" % (located_object.height, self.max_height))
            return True

        # Return before calculating the area if there are no area bounds
        if self.min_area is None and self.max_area is None:
            return False

        # Calculate the area of the object
        area = (
            located_object.get_actual_polygon().area()
            if located_object.has_polygon()
            else located_object.get_actual_rectangle().area()
        )

        # Min area check
        if self.min_area is not None and area < self.min_area:
            self.logger().debug("Area too small: %d < %d" % (area, self.min_area))
            return True

        # Max area check
        if self.max_area is not None and area > self.max_area:
            self.logger().debug("Area too large: %d > %d" % (area, self.max_area))
            return True

        return False

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            ann = LocatedObjects((located_object
                                  for located_object in item.annotation
                                  if not self._should_discard_located_object(located_object)))
            if len(ann) != len(item.annotation):
                item = item.duplicate(annotation=ann)
            result.append(item)

        return flatten_list(result)



import logging
import math
import numpy as np
import statistics
from typing import List, Optional

import shapely
from shapely import Polygon, MultiPolygon, LineString, GeometryCollection, distance
from shapely.geometry.base import BaseGeometry
from shapely.ops import unary_union
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from wai.common.geometry import Point as WaiPoint, Polygon as WaiPolygon

from ._objdet import LABEL_KEY, get_object_label, ObjectDetectionData
from ._imgseg import ImageSegmentationAnnotations

UNION = "union"
INTERSECT = "intersect"
COMBINATIONS = [
    UNION,
    INTERSECT,
]


def locatedobject_bbox_to_shapely(lobj: LocatedObject) -> Polygon:
    """
    Converts the located object rectangle into a shapely Polygon.

    :param lobj: the bbox to convert
    :return: the Polygon
    """
    coords = [
        (lobj.x, lobj.y),
        (lobj.x + lobj.width - 1, lobj.y),
        (lobj.x + lobj.width - 1, lobj.y + lobj.height - 1),
        (lobj.x, lobj.y + lobj.height - 1),
        (lobj.x, lobj.y),
    ]
    return Polygon(coords)


def locatedobject_polygon_to_shapely(lobj: LocatedObject) -> Polygon:
    """
    Converts the located object polygon into a shapely Polygon.

    :param lobj: the polygon to convert
    :return: the Polygon
    """
    if not lobj.has_polygon():
        return locatedobject_bbox_to_shapely(lobj)
    x_list = lobj.get_polygon_x()
    y_list = lobj.get_polygon_y()
    coords = []
    for x, y in zip(x_list, y_list):
        coords.append((x, y))
    coords.append((x_list[0], y_list[0]))
    return Polygon(coords)


def shapely_to_locatedobject(geometry: BaseGeometry, label: str = None) -> LocatedObject:
    """
    Turns the shapely geometry back into a located object.
    Assumes absolute coordinates.

    :param geometry: the geometry to convert
    :type geometry: BaseGeometry
    :param label: the label to set (when not None)
    :type label: str
    :return: the generated object
    :rtype: LocatedObject
    """
    # use convex hull in case of MultiPolygon
    if isinstance(geometry, MultiPolygon):
        geometry = geometry.convex_hull

    minx, miny, maxx, maxy = geometry.bounds
    result = LocatedObject(minx, miny, maxx-minx+1, maxy-miny+1)
    if label is not None:
        result.metadata[LABEL_KEY] = label

    if isinstance(geometry, Polygon):
        x_list, y_list = geometry.exterior.coords.xy
        points = []
        for i in range(len(x_list)):
            points.append(WaiPoint(x=x_list[i], y=y_list[i]))
        result.set_polygon(WaiPolygon(*points))

    return result


def locatedobjects_to_shapely(located_objects: LocatedObjects) -> List[Polygon]:
    """
    Turns the located objects into shapely polygons.

    :param located_objects: the objects to convert
    :type located_objects: LocatedObjects
    :return: the list of polygons
    :rtype: list
    """
    result = []
    for obj in located_objects:
        result.append(locatedobject_polygon_to_shapely(obj))
    return result


def intersect_over_union(poly1: Polygon, poly2: Polygon) -> float:
    """
    Calculates the IoU (intersect over union) for the two polygons.

    :param poly1: the first polygon
    :type poly1: Polygon
    :param poly2: the second polygon
    :type poly2: Polygon
    :return: the IoU
    :type: float
    """
    try:
        intersection = poly2.intersection(poly1)
        if intersection.area > 0:
            union = unary_union([poly2, poly1])
            return intersection.area / union.area
        else:
            return 0
    except:
        print("Failed to compute IoU!")
        return 0


def merge_polygons(combined: ObjectDetectionData, max_slope_diff: float = 1e-6, max_dist: float = 1.0) -> ObjectDetectionData:
    """
    Merges adjacent polygons. Discards metadata apart from score, which it averages across merged objects,
    and the label, which has to be the same across objects.

    :param combined: the input data
    :type combined: ObjectDetectionData
    :param max_slope_diff: the maximum difference between slopes while still being considered parallel
    :type max_slope_diff: float
    :param max_dist: the maximum distance between parallel vertices
    :type max_dist: float
    :return: the (potentially) updated annotations
    :rtype: ObjectDetectionData
    """
    # for each polygon
    #   for each vertex in polygon
    #      compute slope
    # determine parallel vertices between objects
    # compute distance between parallel vertices
    # determine sets of objects to merge

    # determine vertices/slopes/intercepts
    vertices = dict()
    slopes = dict()
    normalized = combined.is_normalized()
    absolute = combined.get_absolute()
    for i in range(len(absolute)):
        vertices[i] = []
        slopes[i] = []
        if absolute[i].has_polygon():
            xs = absolute[i].get_polygon_x()
            ys = absolute[i].get_polygon_y()
        else:
            xs = [absolute[i].x, absolute[i].x + absolute[i].width - 1, absolute[i].x + absolute[i].width - 1, absolute[i].x]
            ys = [absolute[i].y, absolute[i].y, absolute[i].y + absolute[i].height - 1, absolute[i].y + absolute[i].height - 1]
        for n in range(len(xs)):
            # vertex: (x0,y0,x1,y1)
            vertices[i].append(LineString([(xs[n - 1], ys[n - 1]), (xs[n], ys[n])]))
            # slope: m = (y1-y0) / (x1-x0)
            if xs[n] - xs[n - 1] == 0:
                slope = math.inf
            else:
                slope = (ys[n] - ys[n - 1]) / (xs[n] - xs[n - 1])
            slopes[i].append(slope)

    # determine parallel vertices of objects with same label
    parallel = dict()
    for i in range(len(slopes)):
        label_i = get_object_label(absolute[i])
        for n in range(i + 1, len(slopes), 1):
            # only consider objects with the same label
            label_n = get_object_label(absolute[n])
            if label_i != label_n:
                continue
            for i_i in range(len(slopes[i])):
                for n_n in range(len(slopes[n])):
                    is_parallel = False

                    # horizontal lines
                    if (slopes[i][i_i] == 0) and (slopes[n][n_n] == 0):
                        is_parallel = True

                    # vertical lines
                    elif math.isinf(slopes[i][i_i]) and math.isinf(slopes[n][n_n]):
                        is_parallel = True

                    # compare slope
                    else:
                        slope_diff = abs(slopes[i][i_i] - slopes[n][n_n])
                        if slope_diff <= max_slope_diff:
                            is_parallel = True

                    # check distance of parallel vertices
                    if is_parallel:
                        d = distance(vertices[i][i_i], vertices[n][n_n])
                        if d <= max_dist:
                            if i not in parallel:
                                parallel[i] = set()
                            parallel[i].add(n)

    # create sets of objects to merge
    merge_sets = []
    to_merge = set()
    for i, ns in parallel.items():
        all_ = [i, *ns]
        found = None
        for a in all_:
            to_merge.add(a)
            for n, merge_set in enumerate(merge_sets):
                if a in merge_set:
                    found = n
                    break
            if found is not None:
                break
        if found is None:
            merge_sets.append(set(all_))
        else:
            for a in all_:
                merge_sets[found].add(a)

    if len(merge_sets) > 0:
        # transfer all objects that won't get merged
        annotation_new = LocatedObjects()
        for i, obj in enumerate(absolute):
            if i not in to_merge:
                annotation_new.append(obj)

        # merge sets
        for merge_set in merge_sets:
            label = None
            merged = None
            scores = []
            for i in merge_set:
                if label is None:
                    label = get_object_label(absolute[i])
                if "score" in absolute[i].metadata:
                    scores.append(float(absolute[i].metadata["score"]))
                if merged is None:
                    merged = locatedobject_polygon_to_shapely(absolute[i])
                else:
                    merged = shapely.union(merged, locatedobject_polygon_to_shapely(absolute[i]))
            obj = shapely_to_locatedobject(merged, label=label)
            # set average score
            if len(scores) > 0:
                score = statistics.mean(scores)
                obj.metadata["score"] = score
            annotation_new.append(obj)

        # update container
        combined.annotation = annotation_new
        if normalized:
            combined.to_normalized()

    return combined


def fit_located_object(index: int, region: LocatedObject, annotation: LocatedObject, logger: Optional[logging.Logger]) -> LocatedObject:
    """
    Fits the annotation into the specified region, adjusts size if necessary.

    :param index: the index of the current region, gets added to meta-data if >=0
    :type index: int
    :param region: the region object to fit the annotation in
    :type region: LocatedObject
    :param annotation: the annotation to fit
    :type annotation: LocatedObject
    :param logger: the logger to use, can be None
    :type logger: logging.Logger
    :return: the adjusted annotation
    :rtype: LocatedObject
    """
    sregion = locatedobject_bbox_to_shapely(region)
    sbbox = locatedobject_bbox_to_shapely(annotation)
    sintersect = sbbox.intersection(sregion)
    minx, miny, maxx, maxy = [int(x) for x in sintersect.bounds]
    result = LocatedObject(x=minx-region.x, y=miny-region.y, width=maxx-minx+1, height=maxy-miny+1, **annotation.metadata)
    if index > -1:
        result.metadata["region_index"] = index
        result.metadata["region_xywh"] = "%d,%d,%d,%d" % (region.x, region.y, region.width, region.height)

    if annotation.has_polygon():
        spolygon = locatedobject_polygon_to_shapely(annotation)
    else:
        spolygon = locatedobject_bbox_to_shapely(annotation)

    try:
        sintersect = spolygon.intersection(sregion)
    except:
        msg = "Failed to compute intersection!"
        if logger is None:
            print(msg)
        else:
            logger.warning(msg)
        sintersect = None

    if isinstance(sintersect, GeometryCollection):
        for x in sintersect.geoms:
            if isinstance(x, Polygon):
                sintersect = x
                break
    elif isinstance(sintersect, MultiPolygon):
        for x in sintersect.geoms:
            if isinstance(x, Polygon):
                sintersect = x
                break

    if isinstance(sintersect, Polygon):
        x_list, y_list = sintersect.exterior.coords.xy
        points = []
        for i in range(len(x_list)):
            points.append(WaiPoint(x=x_list[i]-region.x, y=y_list[i]-region.y))
        result.set_polygon(WaiPolygon(*points))
    else:
        msg = "Unhandled geometry type returned from intersection, skipping: %s" % str(type(sintersect))
        if logger is None:
            print(msg)
        else:
            logger.warning(msg)

    return result


def fit_layers(region: LocatedObject, annotations: ImageSegmentationAnnotations, suppress_empty: bool) -> ImageSegmentationAnnotations:
    """
    Crops the layers to the region.

    :param region: the region to crop the layers to
    :type region: LocatedObject
    :param annotations: the annotations to crop
    :type annotations: ImageSegmentationAnnotations
    :param suppress_empty: whether to suppress empty annotations
    :type suppress_empty: bool
    :return: the updated annotations
    :rtype: ImageSegmentationAnnotations
    """
    layers = dict()
    for label in annotations.layers:
        layer = annotations.layers[label][region.y:region.y+region.height, region.x:region.x+region.width]
        add = True
        if suppress_empty:
            unique = np.unique(layer)
            # only background? -> skip
            if (len(unique) == 1) and (unique[0] == 0):
                add = False
        if add:
            layers[label] = layer
    return ImageSegmentationAnnotations(annotations.labels[:], layers)



import argparse
import copy
import re

from typing import List

from wai.logging import LOGGING_WARNING
from seppl.io import Filter
from idc.api import ImageClassificationData


class LabelFromName(Filter):
    """
    Extracts the classification label from the image name.
    """

    def __init__(self, regexp: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param regexp: the regular expression to apply to the name (1st group gets used as label)
        :type regexp: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.regexp = regexp

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "label-from-name"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Extracts the classification label from the image name."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageClassificationData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageClassificationData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-r", "--regexp", type=str, help="The regular expression apply to the image name, with the 1st group being used as the label.", default=None, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.regexp = ns.regexp

    def _update(self, data):
        """
        Updates the label of the record.

        :param data: the record to update
        """
        try:
            name = data.image_name
            if name is None:
                self.logger().warning("No image name available: %s" % str(data))
                return data
            m = re.search(self.regexp, name)
            if m is None:
                return data
            label = m.group(1)
            result = copy.deepcopy(data)
            result.annotation = label
            return result
        except:
            self.logger().exception("Failed to extract label from: %s" % data.image_name)
            return data

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        if isinstance(data, list):
            result = list(data)
            for i, item in enumerate(data):
                result[i] = self._update(item)
        else:
            result = self._update(data)

        return result



import argparse
import logging
import os
import sys
import traceback
from typing import List, Optional

from wai.logging import init_logging, set_logging_level, add_logging_level
from idc.core import ENV_IDC_LOGLEVEL
from idc.help import generate_plugin_usage, HELP_FORMATS, HELP_FORMAT_TEXT, HELP_FORMAT_MARKDOWN
from idc.registry import register_plugins, available_pipeline_plugins
from idc.registry import available_readers, available_filters, available_writers, available_generators

HELP = "idc-help"

_logger = logging.getLogger(HELP)

INDEX_TITLE_DEFAULT = "image-dataset-converter plugins"


PLUGIN_TYPE_PIPELINE = "pipeline"
PLUGIN_TYPE_GENERATOR = "generator"
PLUGIN_TYPES = [
    PLUGIN_TYPE_PIPELINE,
    PLUGIN_TYPE_GENERATOR,
]


def _add_plugins_to_index(heading: str, plugins: dict, help_format: str, lines: list):
    """
    Appends a plugin section to the output list.

    :param heading: the heading of the section
    :type heading: str
    :param plugins: the plugins dictionary to add
    :type plugins: dict
    :param help_format: the type of output to generate
    :type help_format: str
    :param lines: the output lines to append the output to
    :type lines: list
    """
    plugin_names = sorted(plugins.keys())
    if len(plugin_names) == 0:
        return
    if help_format == HELP_FORMAT_MARKDOWN:
        lines.append("## " + heading)
        for name in plugin_names:
            lines.append("* [%s](%s.md)" % (name, name))
        lines.append("")
    elif help_format == HELP_FORMAT_TEXT:
        lines.append(heading)
        lines.append("-" * len(heading))
        for name in plugin_names:
            lines.append("- %s" % name)
        lines.append("")
    else:
        raise Exception("Unsupported format for index: %s" % help_format)


def output_help(custom_class_listers: List[str] = None, excluded_class_listers: Optional[List[str]] = None,
                plugin_type: str = None, plugin_name: str = None,
                help_format: str = HELP_FORMAT_TEXT, heading_level: int = 1, output: str = None, index_file: str = None,
                index_title: str = INDEX_TITLE_DEFAULT):
    """
    Generates and outputs the help screen for the plugin.

    :param custom_class_listers: the class listers to generate the entry points for
    :type custom_class_listers: list
    :param excluded_class_listers: the list of class listers to exclude
    :type excluded_class_listers: list
    :param plugin_type: the type of plugin to generate the help for
    :type plugin_type: str
    :param plugin_name: the plugin to generate the help for, None if for all
    :type plugin_name: str
    :param help_format: the format to output
    :type help_format: str
    :param heading_level: the heading level to use (markdown)
    :type heading_level: int
    :param output: the dir/file to save the output to, uses stdout if None
    :type output: str
    :param index_file: the index file to generate in the output directory, ignored if None
    :type index_file: str
    :param index_title: the title to use in the index file
    :type index_title: str
    """
    register_plugins(custom_class_listers=custom_class_listers, excluded_class_listers=excluded_class_listers)

    if help_format not in HELP_FORMATS:
        raise Exception("Unknown help format: %s" % help_format)

    if plugin_type is None:
        plugin_type = PLUGIN_TYPE_PIPELINE
    if plugin_type not in PLUGIN_TYPES:
        raise Exception("Invalid plugin type: %s" % plugin_type)
    available = None
    if plugin_type == PLUGIN_TYPE_PIPELINE:
        available = available_pipeline_plugins()
    elif plugin_type == PLUGIN_TYPE_GENERATOR:
        available = available_generators()
    else:
        raise Exception("Unhandled plugin type: %s" % plugin_type)

    if (plugin_name is None) and ((output is None) or (not os.path.isdir(output))):
        raise Exception("When generating the help for all plugins, the output must be a directory: %s" % output)
    if plugin_name is None:
        plugin_names = available.keys()
    else:
        plugin_names = [plugin_name]
    for p in plugin_names:
        _logger.info("Generating help (%s): %s" % (help_format, p))
        generate_plugin_usage(p, help_format=help_format, heading_level=heading_level, output_path=output)

    if index_file is not None:
        header_lines = []
        if help_format == HELP_FORMAT_MARKDOWN:
            header_lines.append("# %s" % index_title)
            header_lines.append("")
        elif help_format == HELP_FORMAT_TEXT:
            heading = index_title
            header_lines.append(heading)
            header_lines.append("=" * len(heading))
            header_lines.append("")
        else:
            raise Exception("Unsupported format for index: %s" % help_format)

        plugin_lines = []
        if plugin_type == PLUGIN_TYPE_PIPELINE:
            _add_plugins_to_index("Readers", available_readers(), help_format, plugin_lines)
            _add_plugins_to_index("Filters", available_filters(), help_format, plugin_lines)
            _add_plugins_to_index("Writers", available_writers(), help_format, plugin_lines)
        elif plugin_type == PLUGIN_TYPE_GENERATOR:
            _add_plugins_to_index("Generators", available_generators(), help_format, plugin_lines)
        else:
            raise Exception("Unhandled plugin type: %s" % plugin_type)

        if len(plugin_lines) < 0:
            print("No plugins listed, skipping output of index file.")
        else:
            index_file = os.path.join(output, index_file)
            os.makedirs(os.path.dirname(index_file), exist_ok=True)
            with open(index_file, "w") as fp:
                fp.write("\n".join(header_lines))
                fp.write("\n".join(plugin_lines))


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    init_logging(env_var=ENV_IDC_LOGLEVEL)
    parser = argparse.ArgumentParser(
        description="Tool for outputting help for plugins in various formats.",
        prog=HELP,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-c", "--custom_class_listers", metavar="PACKAGE", help="The custom class listers to use, uses the default ones if not provided.", default=None, type=str, required=False, nargs="*")
    parser.add_argument("-e", "--excluded_class_listers", type=str, default=None, help="The comma-separated list of class listers to exclude.", required=False)
    parser.add_argument("-T", "--plugin_type", choices=PLUGIN_TYPES, help="The types of plugins to generate the help for.", default=PLUGIN_TYPE_PIPELINE, type=str, required=False)
    parser.add_argument("-p", "--plugin_name", metavar="NAME", help="The name of the plugin to generate the help for, generates it for all if not specified", default=None, type=str, required=False)
    parser.add_argument("-f", "--help_format", help="The output format to generate", choices=HELP_FORMATS, default=HELP_FORMAT_TEXT, required=False)
    parser.add_argument("-L", "--heading_level", metavar="INT", help="The level to use for the heading", default=1, type=int, required=False)
    parser.add_argument("-o", "--output", metavar="PATH", help="The directory or file to store the help in; outputs it to stdout if not supplied; if pointing to a directory, automatically generates file name from plugin name and help format", type=str, default=None, required=False)
    parser.add_argument("-i", "--index_file", metavar="FILE", help="The file in the output directory to generate with an overview of all plugins, grouped by type (in markdown format, links them to the other generated files)", type=str, default=None, required=False)
    parser.add_argument("-t", "--index_title", metavar="TITLE", help="The title to use in the index file", type=str, default=INDEX_TITLE_DEFAULT, required=False)
    add_logging_level(parser)
    parsed = parser.parse_args(args=args)
    set_logging_level(_logger, parsed.logging_level)
    output_help(custom_class_listers=parsed.custom_class_listers, excluded_class_listers=parsed.excluded_class_listers,
                plugin_type=parsed.plugin_type, plugin_name=parsed.plugin_name, help_format=parsed.help_format,
                heading_level=parsed.heading_level, output=parsed.output,
                index_file=parsed.index_file, index_title=parsed.index_title)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        print("options: %s" % str(sys.argv[1:]), file=sys.stderr)
        return 1


if __name__ == '__main__':
    main()



import argparse
from typing import List

from seppl.io import Filter
from wai.logging import LOGGING_WARNING
from idc.api import ObjectDetectionData, ImageClassificationData, flatten_list, make_list, get_object_label


MULTIPLICITY_ERROR = "error"
MULTIPLICITY_MAJORITY = "majority"
MULTIPLICITY_SINGLE = "single"
MULTIPLICITY_SKIP = "skip"
MULTIPLICITY = [
    MULTIPLICITY_ERROR,
    MULTIPLICITY_MAJORITY,
    MULTIPLICITY_SINGLE,
    MULTIPLICITY_SKIP,
]


class ObjectDetectionToImageClassification(Filter):
    """
    Converts object detection annotations into image classification ones.
    """

    def __init__(self, multiplicity: str = MULTIPLICITY_ERROR,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param multiplicity: how to handle instances with more than one located object
        :type multiplicity: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.multiplicity = multiplicity

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "od-to-ic"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Converts object detection annotations into image classification ones."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageClassificationData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-m", "--multiplicity", choices=MULTIPLICITY, default=MULTIPLICITY_ERROR, help="How to handle instances with more than one located object", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.multiplicity = ns.multiplicity

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            ann = item.annotation
            if (ann is None) or (len(ann) == 0):
                item_new = ImageClassificationData(source=item.source, image=item.image, data=item.data, metadata=item.get_metadata())
            else:
                if len(ann) > 1:
                    if self.multiplicity == "error":
                        raise Exception("More than one detected object for: %s" % item.image_name)
                    elif self.multiplicity == "skip":
                        return
                    elif self.multiplicity == "single":
                        labels = set(map(get_object_label, ann))
                        if len(labels) > 1:
                            raise Exception("More than one type of detected object for: %s" % item.image_name)
                        label = labels.pop()
                    elif self.multiplicity == "majority":
                        labels = {}
                        for label in map(get_object_label, ann):
                            if label in labels:
                                labels[label] += 1
                            else:
                                labels[label] = 1
                        label = max(labels.keys(), key=labels.get)
                    else:
                        raise Exception("Unhandled multiplicity: %s" % self.multiplicity)
                else:
                    label = get_object_label(ann[0])
                item_new = ImageClassificationData(source=item.source, image=item.image, data=item.data, metadata=item.get_metadata(), annotation=label)

            result.append(item_new)

        return flatten_list(result)



from typing import List

from seppl import AnyData
from seppl.io import Filter
from wai.logging import LOGGING_WARNING

from idc.api import ImageData, flatten_list, make_list


class DiscardNegatives(Filter):
    """
    Discards negative images, i.e., ones without annotations.
    """

    def __init__(self, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.kept = 0
        self.discarded = 0

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "discard-negatives"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Discards negative images, i.e., ones without annotations."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self.kept = 0
        self.discarded = 0

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            if item.has_annotation():
                self.kept += 1
                result.append(item)
            else:
                self.discarded += 1
                self.logger().info("Discarding image: %s" % item.image_name)

        return flatten_list(result)

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        self.logger().info("# kept: %d" % self.kept)
        self.logger().info("# discarded: %d" % self.discarded)



from ._bluechannel import BlueChannelImageSegmentationWriter
from ._grayscale import GrayscaleImageSegmentationWriter
from ._indexedpng import IndexedPngImageSegmentationWriter
from ._layersegments import LayerSegmentsImageSegmentationWriter



import copy
from typing import Tuple, Dict, Union, Optional

from PIL import Image
from wai.common.adams.imaging.locateobjects import LocatedObject, LocatedObjects
from wai.common.adams.imaging.locateobjects import NormalizedLocatedObjects, NormalizedLocatedObject
from wai.common.adams.imaging.locateobjects import absolute_to_normalized, normalized_to_absolute

from ._data import ImageData

DEFAULT_LABEL = "object"

LABEL_KEY = "type"


def get_object_label(located_object: Union[LocatedObject, NormalizedLocatedObject], default_label: Optional[str] = DEFAULT_LABEL) -> Optional[str]:
    """
    Returns the object label from the located object.

    :param located_object: the located object to get the label from the meta-data
    :type located_object: LocatedObject
    :param default_label: the default label to use in case no label is stored in meta-data
    :type default_label: str
    :return: the label
    :rtype: str
    """
    if LABEL_KEY in located_object.metadata:
        return str(located_object.metadata[LABEL_KEY])
    else:
        return default_label


def set_object_label(located_object: Union[LocatedObject, NormalizedLocatedObject], label: str):
    """
    Sets the object label for the located object.

    :param located_object: the located object to set the label for in its meta-data
    :type located_object: LocatedObject
    :param label: the label to set
    :type label: str
    """
    located_object.metadata[LABEL_KEY] = label


class ObjectDetectionData(ImageData):
    """
    The annotations are LocatedObjects or NormalizedLocatedObjects.
    """

    def __init__(self, source: str = None, image_name: str = None, data: bytes = None,
                 image: Image.Image = None, image_format: str = None, image_size: Tuple[int, int] = None,
                 metadata: Dict = None, annotation: Union[LocatedObjects, NormalizedLocatedObjects] = None):

        super().__init__(source=source, image_name=image_name, data=data,
                         image=image, image_format=image_format, image_size=image_size,
                         metadata=metadata, annotation=annotation)

    def has_annotation(self) -> bool:
        """
        Checks whether annotations are present.

        :return: True if annotations present
        :rtype: bool
        """
        return (self.annotation is not None) and (len(self.annotation) > 0)

    def is_normalized(self) -> bool:
        """
        Returns whether the annotations are normalized or absolute.

        :return: True if normalized
        :rtype: bool
        """
        return isinstance(self.annotation, NormalizedLocatedObjects)

    def get_normalized(self) -> Optional[NormalizedLocatedObjects]:
        """
        Returns normalized annotations.
        """
        if self.annotation is None:
            return None
        if isinstance(self.annotation, NormalizedLocatedObjects):
            return self.annotation
        if self.image_size is None:
            return None
        if isinstance(self.annotation, LocatedObjects):
            width, height = self.image_size
            return absolute_to_normalized(self.annotation, width, height)
        raise Exception("Unhandled type of annotations: %s" % str(type(self.annotation)))

    def to_normalized(self):
        """
        Turns the absolute annotations into normalized ones (in-place).
        """
        if self.is_normalized():
            return
        self.annotation = self.get_normalized()

    def get_absolute(self) -> Optional[LocatedObjects]:
        """
        Returns absolute annotations.

        :return: the absolute annotations, None if not available
        :rtype: LocatedObjects
        """
        if self.annotation is None:
            return None
        if isinstance(self.annotation, LocatedObjects):
            return self.annotation
        if self.image_size is None:
            return None
        if isinstance(self.annotation, NormalizedLocatedObjects):
            width, height = self.image_size
            return normalized_to_absolute(self.annotation, width, height)
        raise Exception("Unhandled type of annotations: %s" % str(type(self.annotation)))

    def to_absolute(self):
        """
        Turns the normalized annotations into absolute ones (in-place).
        """
        if not self.is_normalized():
            return
        self.annotation = self.get_absolute()

    def _annotation_to_dict(self):
        """
        Turns the annotations into a dictionary.

        :return: the generated dictionary
        :rtype: dict
        """
        objs = []
        for lobj in self.annotation:
            obj = dict()
            obj["x"] = lobj.x
            obj["y"] = lobj.y
            obj["width"] = lobj.width
            obj["height"] = lobj.height
            obj["metadata"] = copy.deepcopy(lobj.metadata)
            if lobj.has_polygon():
                obj["poly_x"] = lobj.get_polygon_x()
                obj["poly_y"] = lobj.get_polygon_y()
            objs.append(obj)

        result = dict()
        result["objects"] = objs
        result["normalized"] = self.is_normalized()

        return result



import argparse
import os
from typing import List

from seppl.placeholders import PlaceholderSupporter, placeholder_list, expand_placeholders
from seppl.io import Filter
from wai.logging import LOGGING_WARNING

from idc.api import ObjectDetectionData, ImageClassificationData, ImageSegmentationData, get_object_label, make_list

OUTPUT_FORMAT_TEXT = "text"
OUTPUT_FORMAT_COMMASEP = "comma-separated"
OUTPUT_FORMAT_CUSTOMSEP = "custom-separator"
OUTPUT_FORMATS = [
    OUTPUT_FORMAT_TEXT,
    OUTPUT_FORMAT_COMMASEP,
    OUTPUT_FORMAT_CUSTOMSEP,
]


class WriteLabels(Filter, PlaceholderSupporter):
    """
    Collects labels passing through and writes them to the specified file (stdout if not provided).
    """

    def __init__(self, output_file: str = None, output_format: str = OUTPUT_FORMAT_TEXT, custom_sep: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param output_file: the file to write the labels to
        :type output_file: str
        :param output_format: the format to use
        :type output_format: str
        :param custom_sep: the custom separator to use (if OUTPUT_FORMAT_CUSTOMSEP)
        :type custom_sep: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.output_file = output_file
        self.output_format = output_format
        self.custom_sep = custom_sep
        self._labels = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "write-labels"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Collects labels passing through and writes them to the specified file (stdout if not provided)."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData, ImageClassificationData, ImageSegmentationData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData, ImageClassificationData, ImageSegmentationData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output_file", type=str, default=None, help="The file to write the labels to; uses stdout if not provided; " + placeholder_list(obj=self), required=False)
        parser.add_argument("-f", "--output_format", choices=OUTPUT_FORMATS, default=OUTPUT_FORMAT_TEXT, help="The format to use for the labels", required=False)
        parser.add_argument("-s", "--custom_sep", default=None, help="The custom separator to use; use \\t, \\n or \\r for tab, new line or carriage return", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_file = ns.output_file
        self.output_format = ns.output_format
        self.custom_sep = ns.custom_sep

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._labels = set()
        if self.output_format is None:
            self.output_format = OUTPUT_FORMAT_TEXT
        if self.output_format == OUTPUT_FORMAT_CUSTOMSEP:
            if self.custom_sep is None:
                raise Exception("No custom separator specified!")
            self.custom_sep = self.custom_sep.replace("\\t", "\t").replace("\\n", "\n").replace("\\r", "\r")

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        for item in make_list(data):
            if not item.has_annotation():
                continue

            if isinstance(item, ImageClassificationData):
                self._labels.add(item.annotation)
            elif isinstance(item, ObjectDetectionData):
                for obj in item.annotation:
                    label = get_object_label(obj)
                    if label is not None:
                        self._labels.add(label)
            elif isinstance(item, ImageSegmentationData):
                for label in item.annotation.layers:
                    self._labels.add(label)

        return data

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()

        labels = sorted(self._labels)
        if self.output_format == OUTPUT_FORMAT_TEXT:
            sep = "\n"
        elif self.output_format == OUTPUT_FORMAT_COMMASEP:
            sep = ","
        elif self.output_format == OUTPUT_FORMAT_CUSTOMSEP:
            sep = self.custom_sep
        else:
            raise Exception("Unhandled output format: %s" % self.output_format)
        text = sep.join(labels)

        output_file = None
        if self.output_file is not None:
            output_file = expand_placeholders(self.output_file)
        if (output_file is None) or os.path.isdir(output_file):
            print(text)
        else:
            self.logger().info("Writing labels to: %s" % output_file)
            with open(output_file, "w") as fp:
                fp.write(text)
                fp.write("\n")



from PIL import Image
from typing import Tuple, Dict

from ._data import ImageData


class ImageClassificationData(ImageData):
    """
    The annotations are the classification label.
    """

    def __init__(self, source: str = None, image_name: str = None, data: bytes = None,
                 image: Image.Image = None, image_format: str = None, image_size: Tuple[int, int] = None,
                 metadata: Dict = None, annotation: str = None):

        super().__init__(source=source, image_name=image_name, data=data,
                         image=image, image_format=image_format, image_size=image_size,
                         metadata=metadata, annotation=annotation)

    def has_annotation(self) -> bool:
        """
        Checks whether annotations are present.

        :return: True if annotations present
        :rtype: bool
        """
        return (self.annotation is not None) and (len(self.annotation) > 0)

    def _annotation_to_dict(self):
        """
        Turns the annotations into a dictionary.

        :return: the generated dictionary
        :rtype: dict
        """
        return {"label": self.annotation}



import traceback
from typing import Tuple

from matplotlib import font_manager
from PIL import ImageFont, Image, ImageDraw


DEFAULT_FONT_FAMILY = "sans\\-serif"

TEXT_SIZE_OLD = None


def load_font(logger, family, size):
    """
    Attempts to instantiate the specified font family.

    :param logger: the logger instance to use, ignored if None
    :param family: the TTF font family
    :type family: str
    :param size: the size to use
    :type size: int
    :return: the Pillow font
    """
    try:
        mpl_font = font_manager.FontProperties(family=family)
        font_file = font_manager.findfont(mpl_font)
        return ImageFont.truetype(font_file, size)
    except:
        msg = "Failed to instantiate font family '%s', falling back on '%s'" % (family, DEFAULT_FONT_FAMILY)
        if logger is not None:
            logger.warning(msg, exc_info=True)
        else:
            print(msg)
            print(traceback.format_exc())

        mpl_font = font_manager.FontProperties(family=DEFAULT_FONT_FAMILY)
        font_file = font_manager.findfont(mpl_font)
        return ImageFont.truetype(font_file, size)


def _text_size_old(text: str, font) -> Tuple[int, int]:
    """
    Computes the width/height of a text using the specified Pillow font.
    Uses ImageDraw.textsize.

    :param text: the text to get the dimensions for
    :type text: str
    :param font: the Pillow font to use
    :return: the width and height
    :rtype: tuple
    """
    im = Image.new(mode="P", size=(0, 0))
    draw = ImageDraw.Draw(im)
    w, h = draw.textsize(text, font=font)
    return w, h


def _text_size_new(text: str, font) -> Tuple[int, int]:
    """
    Computes the width/height of a text using the specified Pillow font.
    Newer versions of Pillow deprecated ImageDraw.textsize
    https://levelup.gitconnected.com/how-to-properly-calculate-text-size-in-pil-images-17a2cc6f51fd

    :param text: the text to get the dimensions for
    :type text: str
    :param font: the Pillow font to use
    :return: the width and height
    :rtype: tuple
    """
    ascent, descent = font.getmetrics()
    w = font.getmask(text).getbbox()[2]
    h = font.getmask(text).getbbox()[3] + descent
    return w, h


def text_size(text: str, font) -> Tuple[int, int]:
    """
    Computes the width/height of a text using the specified Pillow font.

    :param text: the text to get the dimensions for
    :type text: str
    :param font: the Pillow font to use
    :return: the width and height
    :rtype: tuple
    """
    global TEXT_SIZE_OLD

    if TEXT_SIZE_OLD is None:
        try:
            w, h = _text_size_old(text, font)
            TEXT_SIZE_OLD = True
        except:
            w, h = _text_size_new(text, font)
            TEXT_SIZE_OLD = False
    else:
        if TEXT_SIZE_OLD:
            w, h = _text_size_old(text, font)
        else:
            w, h = _text_size_new(text, font)
    return w, h



import argparse
import numpy as np
import re
from typing import List

from PIL import Image, ImageDraw
from seppl.io import Filter
from wai.logging import LOGGING_WARNING
from idc.api import ObjectDetectionData, ImageSegmentationData, ImageSegmentationAnnotations, flatten_list, make_list, get_object_label


class ObjectDetectionToImageSegmentation(Filter):
    """
    Converts object detection annotations into image segmentation ones.
    """

    def __init__(self, labels: List[str] = None, regexp: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param labels: the labels to use
        :type labels: list
        :param regexp: the regular expression for using only a subset
        :type regexp: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.labels = labels
        self.regexp = regexp
        self._pattern = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "od-to-is"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Converts object detection annotations into image segmentation ones."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageSegmentationData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("--labels", type=str, default=None, help="The labels to use", required=False, nargs="*")
        parser.add_argument("--regexp", type=str, default=None, help="Regular expression for using only a subset of labels", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.labels = ns.labels
        self.regexp = ns.regexp

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._pattern = re.compile(self.regexp) if (self.regexp is not None) else None
        if (self.labels is None) or (len(self.labels) == 0) and (self._pattern is None):
            raise Exception("No labels/layers or regexp defined!")

    def _label_matches(self, label: str) -> bool:
        """
        Checks whether the label matches.

        :param label:   The label to test.
        :return:        True if the label matches, false if not.
        """
        if len(self.labels) == 0 and self._pattern is None:
            return False
        elif len(self.labels) == 0:
            return bool(self._pattern.match(label))
        elif self._pattern is None:
            return label in self.labels
        else:
            return bool(self._pattern.match(label)) or label in self.labels

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            # draw bboxes/polygons
            imgs = dict()
            draws = dict()
            absolute = item.get_absolute()
            for obj in absolute:
                label = get_object_label(obj)
                if not self._label_matches(label):
                    continue
                if label not in imgs:
                    img = Image.new("1", item.image_size)
                    imgs[label] = img
                    draw = ImageDraw.Draw(img)
                    draws[label] = draw
                if obj.has_polygon():
                    points = [(x, y) for x, y in zip(obj.get_polygon_x(), obj.get_polygon_y())]
                else:
                    points = [
                        (obj.x, obj.y),
                        (obj.x + obj.width + 1, obj.y),
                        (obj.x + obj.width + 1, obj.y + obj.height + 1),
                        (obj.x, obj.y + obj.height + 1)]
                draws[label].polygon(points, fill=255, outline=255, width=1)

            # generate layers
            layers = dict()
            for label in imgs:
                layers[label] = np.asarray(imgs[label]).astype(np.uint8)
                layers[label] = np.where(layers[label] == 1, 255, 0).astype(np.uint8)

            self.logger().info("Layers generated: %s" % ",".join(sorted(layers.keys())))

            # generate imgseg container
            ann = ImageSegmentationAnnotations(labels=self.labels[:], layers=layers)
            imgseg = ImageSegmentationData(source=item.source, image_name=item.image_name, data=item.data, annotation=ann, metadata=item.get_metadata())
            result.append(imgseg)

        return flatten_list(result)



import argparse
from typing import List

from seppl.io import Filter
from wai.logging import LOGGING_WARNING

from idc.api import flatten_list, make_list, ImageData, DATATYPES, data_type_to_class, load_function


class PythonFunctionFilter(Filter):
    """
    Selects a sub-sample from the stream.
    """

    def __init__(self, function: str = None, input_data_type: str = None, output_data_type: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param function: the function to use (module_name:function_name)
        :type function: str
        :param input_data_type: the type of output to generate from the images
        :type input_data_type: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.function = function
        self.input_data_type = input_data_type
        self.output_data_type = output_data_type
        self._input_cls = None
        self._output_cls = None
        self._function = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "pyfunc-filter"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "The declared Python function processes image containers of the specified input type and generates ones of the specified output type. The function must handle a single image container or an iterable of image containers and return a single image container or an iterable of image containers."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input_data_type", choices=DATATYPES, type=str, default=None, help="The type of data to receive", required=True)
        parser.add_argument("-f", "--function", type=str, default=None, help="The Python function to use, format: module_name:function_name", required=True)
        parser.add_argument("-o", "--output_data_type", choices=DATATYPES, type=str, default=None, help="The type of data to forward", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.input_data_type = ns.input_data_type
        self.function = ns.function
        self.output_data_type = ns.output_data_type

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._input_cls = data_type_to_class(self.input_data_type)
        self._function = load_function(self.function)
        self._output_cls = data_type_to_class(self.output_data_type)

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            if issubclass(type(item), self._input_cls):
                new_items = self._function(item)
                if new_items is None:
                    continue
                new_items = make_list(new_items)
                for new_item in new_items:
                    if new_item is None:
                        continue
                    if issubclass(type(item), self._output_cls):
                        result.append(new_item)
                    else:
                        self.logger().error("Function '%s' did not return an object of type '%s' but of type '%s'!" % (self.function, self._output_cls, type(item)))
            else:
                self.logger().error("Did not receive an object of type '%s' but of type '%s'!" % (self._input_cls, type(item)))

        return flatten_list(result)



import argparse
from typing import List

import numpy as np
from PIL import Image
from seppl import AnyData
from seppl.io import Filter
from wai.logging import LOGGING_WARNING

from idc.api import ImageData, flatten_list, make_list, array_to_image, safe_deepcopy


SORTING_COLS = "cols"
SORTING_ROWS = "rows"
SORTING_COLS_THEN_ROWS = "cols-then-rows"
SORTING_ROWS_THEN_COLS = "rows-then-cols"
SORTING_TYPES = [
    SORTING_COLS,
    SORTING_ROWS,
    SORTING_COLS_THEN_ROWS,
    SORTING_ROWS_THEN_COLS,
]


class SortPixels(Filter):
    """
    Sorts the (grayscale) pixels in ascending order per row.
    """

    def __init__(self, sorting: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param sorting: the type of sorting to use
        :type sorting: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.sorting = sorting

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "sort-pixels"

    def description(self) -> str:
        """
        Returns a description of the filter.

        :return: the description
        :rtype: str
        """
        return "Sorts the (grayscale) pixels in ascending order per row."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-s", "--sorting", choices=SORTING_TYPES, help="How to sort the pixels.", default=SORTING_COLS, required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.sorting = ns.sorting

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.sorting is None:
            self.sorting = SORTING_COLS
        if self.sorting not in SORTING_TYPES:
            raise Exception("Unsupported sorting type (%s): %s" % ("|".join(SORTING_TYPES), self.sorting))

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []
        for item in make_list(data):
            orig_array = np.array(item.image)
            if len(orig_array.shape) > 2:
                self.logger().warning("Expected grayscale image, but got these dimensions: %s" % str(orig_array.shape))
                result.append(item)
            else:
                if self.sorting == SORTING_COLS:
                    sorted_array = np.sort(orig_array)
                elif self.sorting == SORTING_ROWS:
                    sorted_array = np.sort(orig_array.transpose())
                elif self.sorting == SORTING_COLS_THEN_ROWS:
                    sorted_array = np.sort(orig_array)
                    sorted_array = np.sort(sorted_array.transpose())
                    sorted_array = sorted_array.transpose()
                elif self.sorting == SORTING_ROWS_THEN_COLS:
                    sorted_array = np.sort(orig_array.transpose())
                    sorted_array = np.sort(sorted_array.transpose())
                else:
                    raise Exception("Unhandled sorting type: %s" % self.sorting)
                sorted_img = Image.fromarray(np.uint8(sorted_array), mode='L')
                sorted_item = type(item)(source=None, image_name=item.image_name,
                                         data=array_to_image(sorted_array, item.image_format),
                                         image=sorted_img, image_format=item.image_format,
                                         metadata=safe_deepcopy(item.get_metadata()),
                                         annotation=safe_deepcopy(item.annotation))
                result.append(sorted_item)

        return flatten_list(result)



import argparse
import csv
import os
from typing import List

from wai.common.adams.imaging.locateobjects import absolute_to_normalized
from wai.logging import LOGGING_WARNING

from idc.api import ObjectDetectionData, SplittableStreamWriter, make_list, AnnotationsOnlyWriter, add_annotations_only_param
from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter


class ROIObjectDetectionWriter(SplittableStreamWriter, AnnotationsOnlyWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_dir: str = None, suffix: str = "-rois.csv", size_mode: bool = False,
                 annotations_only: bool = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param output_dir: the output directory to save the image/csv in
        :type output_dir: str
        :param suffix: the suffix of the ROI CSV files (eg -rois.csv)
        :type suffix: str
        :param size_mode: whether to output w/h rather than x1/y1
        :type size_mode: bool
        :param annotations_only: whether to output only the annotations and not the images
        :type annotations_only: bool
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.suffix = suffix
        self.size_mode = size_mode
        self.annotations_only = annotations_only
        self._label_mapping = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-roicsv-od"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the bounding box/polygon definitions in a ROI .csv file alongside the image."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the images/.csv files in. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        parser.add_argument("-s", "--suffix", metavar="SUFFIX", type=str, default="-rois.csv", help="The suffix used by the ROI CSV files.", required=False)
        parser.add_argument("--size_mode", action="store_true", help="Whether to output w/h rather than x1/y1.", required=False)
        add_annotations_only_param(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.suffix = ns.suffix
        self.size_mode = ns.size_mode
        self.annotations_only = ns.annotations_only

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.annotations_only is None:
            self.annotations_only = False
        self._label_mapping = dict()

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            sub_dir = self.session.expand_placeholders(self.output_dir)
            if self.splitter is not None:
                split = self.splitter.next(item=item.image_name)
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating dir: %s" % sub_dir)
                os.makedirs(sub_dir)

            path = os.path.join(sub_dir, item.image_name)
            if not self.annotations_only:
                self.logger().info("Writing image to: %s" % path)
                item.save_image(path)

            if item.has_annotation():
                path = os.path.splitext(path)[0] + self.suffix
                self.logger().info("Writing ROI CSV to: %s" % path)
                aobjs = item.annotation
                width, height = item.image_size
                nobjs = absolute_to_normalized(aobjs, width, height)
                if self.size_mode:
                    fields = ["file", "x", "y", "w", "h", "xn", "yn", "wn", "hn", "label", "label_str", "score", "poly_x", "poly_y", "poly_xn", "poly_yn", "minrect_w", "minrect_h"]
                else:
                    fields = ["file", "x0", "y0", "x1", "y1", "x0n", "y0n", "x1n", "y1n", "label", "label_str", "score", "poly_x", "poly_y", "poly_xn", "poly_yn", "minrect_w", "minrect_h"]
                with open(path, "w") as fp:
                    writer = csv.DictWriter(fp, fields)
                    writer.writeheader()
                    for aobj, nobj in zip(aobjs, nobjs):
                        if "type" in aobj.metadata:
                            label_str = aobj.metadata["type"]
                            if label_str not in self._label_mapping:
                                self._label_mapping[label_str] = len(self._label_mapping)
                            label = self._label_mapping[label_str]
                        else:
                            label_str = ""
                            label = ""
                        row = dict()
                        row["file"] = item.image_name
                        if self.size_mode:
                            row["x"] = aobj.x
                            row["y"] = aobj.y
                            row["h"] = aobj.width
                            row["y"] = aobj.height
                            row["xn"] = nobj.x
                            row["yn"] = nobj.y
                            row["wn"] = nobj.width
                            row["hn"] = nobj.height
                        else:
                            row["x0"] = aobj.x
                            row["y0"] = aobj.y
                            row["x1"] = aobj.x + aobj.width - 1
                            row["y1"] = aobj.y + aobj.height - 1
                            row["x0n"] = nobj.x
                            row["y0n"] = nobj.y
                            row["x1n"] = nobj.x + nobj.width
                            row["y1n"] = nobj.y + nobj.height
                        row["label"] = label
                        row["label_str"] = label_str
                        if "score" in aobj.metadata:
                            row["score"] = aobj.metadata["score"]
                        if aobj.has_polygon():
                            row["poly_x"] = ",".join([str(x) for x in aobj.get_polygon_x()])
                            row["poly_y"] = ",".join([str(y) for y in aobj.get_polygon_y()])
                            row["poly_xn"] = ",".join(["%.6f" % x for x in nobj.get_polygon_x()])
                            row["poly_yn"] = ",".join(["%.6f" % y for y in nobj.get_polygon_y()])
                        if "minrect_w" in aobj.metadata:
                            row["minrect_w"] = aobj.metadata["minrect_w"]
                        if "minrect_h" in aobj.metadata:
                            row["minrect_h"] = aobj.metadata["minrect_h"]
                        writer.writerow(row)



import argparse
from typing import List

from wai.logging import LOGGING_WARNING
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from seppl.io import Filter
from idc.api import ObjectDetectionData, flatten_list, make_list


class PolygonDiscarder(Filter):
    """
    Removes polygons that fall outside the specified point limits (skips annotations with no polygons).
    """

    def __init__(self, min_points: int = None, max_points: int = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param min_points: the minimum number of points in the polygon, ignored if None
        :type min_points: int
        :param max_points: the maximum number of points in the polygon, ignored if None
        :type max_points: int
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.min_points = min_points
        self.max_points = max_points

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "polygon-discarder"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Removes polygons that fall outside the specified point limits (skips annotations with no polygons)."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-m", "--min_points", type=int, default=None, help="The minimum number of points", required=False)
        parser.add_argument("-M", "--max_points", type=int, default=None, help="The maximum number of points", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.min_points = ns.min_points
        self.max_points = ns.max_points

    def _should_discard_located_object(self, located_object: LocatedObject) -> bool:
        """
        Decides if the located object should be discarded.

        :param located_object:  The located object.
        :return:                True if it should be discarded,
                                False if it should be kept.
        """
        if not located_object.has_polygon():
            return False

        poly_x = located_object.get_polygon_x()

        # Min points
        if self.min_points is not None and len(poly_x) < self.min_points:
            self.logger().debug("Too few points: %d < %d" % (len(poly_x), self.min_points))
            return True

        # Max points
        if self.max_points is not None and len(poly_x) > self.max_points:
            self.logger().debug("Too many points: %d > %d" % (len(poly_x), self.max_points))
            return True

        return False

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            ann = LocatedObjects((located_object
                                  for located_object in item.annotation
                                  if not self._should_discard_located_object(located_object)))
            if len(ann) != len(item.annotation):
                item = item.duplicate(annotation=ann)
            result.append(item)

        return flatten_list(result)



import argparse
import logging
import os.path
import sys
import traceback

from typing import List, Tuple, Optional, Dict

from seppl import enumerate_plugins, is_help_requested, split_args, args_to_objects, Plugin, check_compatibility
from seppl.placeholders import load_user_defined_placeholders
from seppl.io import execute, Reader, Filter, MultiFilter, Writer
from wai.logging import init_logging, set_logging_level, add_logging_level, LOGGING_LEVELS

from idc.core import Session, ENV_IDC_LOGLEVEL
from idc.help import generate_plugin_usage
from idc.registry import available_readers, available_filters, available_writers, REGISTRY


CONVERT = "idc-convert"


DEFAULT_UPDATE_INTERVAL = 1000


def _available_plugins() -> Dict[str, Plugin]:
    """
    Returns the available reader/filter/writer plugins.

    :return: the dictionary of plugins (name/Plugin)
    :rtype: dict
    """
    result = dict()
    result.update(available_readers())
    result.update(available_filters())
    result.update(available_writers())
    return result


def _print_usage(plugin_details: bool = False):
    """
    Prints the program usage to stdout.
    Ensure global options are in sync with parser in parse_args method below.

    :param plugin_details: whether to output the plugin details as well
    :type plugin_details: bool
    """
    cmd = "usage: " + CONVERT
    prefix = " " * (len(cmd) + 1)
    logging_levels = ",".join(LOGGING_LEVELS)
    print(cmd + " [-h|--help|--help-all|--help-plugin NAME]")
    print(prefix + "[-u INTERVAL] [-b|--force_batch] [--placeholders FILE]")
    print(prefix + "[-l {%s}]" % logging_levels)
    print(prefix + "reader")
    print(prefix + "[filter [filter [...]]]")
    print(prefix + "[writer]")
    print()
    print("Tool for converting between image annotation dataset formats.")
    print()
    print("readers (%d):\n" % len(available_readers()) + enumerate_plugins(available_readers().keys(), aliases=REGISTRY.all_aliases, prefix="   "))
    print("filters (%d):\n" % len(available_filters()) + enumerate_plugins(available_filters().keys(), aliases=REGISTRY.all_aliases, prefix="   "))
    print("writers (%d):\n" % len(available_writers()) + enumerate_plugins(available_writers().keys(), aliases=REGISTRY.all_aliases, prefix="   "))
    print()
    print("optional arguments:")
    print("  -h, --help            show basic help message and exit")
    print("  --help-all            show basic help message plus help on all plugins and exit")
    print("  --help-plugin NAME    show help message for plugin NAME and exit")
    print("  -u INTERVAL, --update_interval INTERVAL")
    print("                        outputs the progress every INTERVAL records (default: %d)" % DEFAULT_UPDATE_INTERVAL)
    print("  -l {%s}, --logging_level {%s}" % (logging_levels, logging_levels))
    print("                        the logging level to use (default: WARN)")
    print("  -b, --force_batch     processes the data in batches")
    print("  --placeholders FILE")
    print("                        The file with custom placeholders to load (format: key=value).")
    print()
    if plugin_details:
        for plugin in sorted(_available_plugins().keys()):
            generate_plugin_usage(plugin)


def _parse_args(args: List[str], require_reader: bool = True, require_writer: bool = True) -> Tuple[Optional[Reader], Optional[Filter], Optional[Writer], Session]:
    """
    Parses the arguments.

    :param args: the arguments to parse
    :type args: list
    :param require_reader: whether a reader is required
    :type require_reader: bool
    :param require_writer: whether a writer is required
    :type require_writer: bool
    :return: tuple of (reader, filter, writer, session), the filter can be None
    :rtype: tuple
    """
    # help requested?
    help_requested, plugin_details, plugin_name = is_help_requested(args)
    if help_requested:
        if plugin_name is not None:
            generate_plugin_usage(plugin_name)
        else:
            _print_usage(plugin_details=plugin_details)
        sys.exit(0)

    parsed = split_args(args, list(_available_plugins().keys()))
    plugins = args_to_objects(parsed, _available_plugins(), allow_global_options=True)
    reader = None
    writer = None
    filters = []
    for plugin in plugins:
        if isinstance(plugin, Reader):
            if reader is None:
                reader = plugin
                continue
            else:
                raise Exception("Only one reader can be defined!")

        if isinstance(plugin, Filter):
            filters.append(plugin)
            continue

        if isinstance(plugin, Writer):
            if writer is None:
                writer = plugin
                continue
            else:
                raise Exception("Only one writer can be defined!")

        raise Exception("Unhandled plugin type: %s" % str(type(plugin)))

    # checks whether valid pipeline
    if (reader is None) and require_reader:
        raise Exception("No reader defined!")
    if (writer is None) and require_writer:
        raise Exception("No writer defined!")
    if len(filters) == 0:
        filter_ = None
    elif len(filters) == 1:
        filter_ = filters[0]
    else:
        filter_ = MultiFilter(filters=filters)

    # check compatibility
    if writer is not None:
        if filter_ is not None:
            check_compatibility([reader, filter_, writer])
        else:
            check_compatibility([reader, writer])

    # global options?
    # see print_usage() method above
    parser = argparse.ArgumentParser()
    add_logging_level(parser)
    parser.add_argument("-u", "--update_interval", type=int, default=DEFAULT_UPDATE_INTERVAL)
    parser.add_argument("-b", "--force_batch", action="store_true")
    parser.add_argument("--placeholders")
    session = Session(options=parser.parse_args(parsed[""] if ("" in parsed) else []))
    session.logger = logging.getLogger(CONVERT)
    set_logging_level(session.logger, session.options.logging_level)
    if session.options.placeholders is not None:
        if not os.path.exists(session.options.placeholders):
            session.logger.error("Placeholder file not found: %s" % session.options.placeholders)
        else:
            session.logger.info("Loading custom placeholders from: %s" % session.options.placeholders)
            load_user_defined_placeholders(session.options.placeholders)

    return reader, filter_, writer, session


def main(args=None):
    """
    The main method for parsing command-line arguments.

    :param args: the commandline arguments, uses sys.argv if not supplied
    :type args: list
    """
    init_logging(env_var=ENV_IDC_LOGLEVEL)
    _args = sys.argv[1:] if (args is None) else args
    try:
        reader, filter_, writer, session = _parse_args(_args, require_writer=False)
    except Exception:
        traceback.print_exc()
        print("options: %s" % str(_args), file=sys.stderr)
        _print_usage()
        sys.exit(1)

    session.logger.info("options: %s" % str(_args))

    execute(reader, filter_, writer, session)


def sys_main() -> int:
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    """
    try:
        main()
        return 0
    except Exception:
        traceback.print_exc()
        return 1


if __name__ == '__main__':
    main()



import argparse
from typing import List

from wai.logging import LOGGING_WARNING
from seppl import split_args, split_cmdline, Plugin, AnyData, Initializable
from seppl.io import Writer, BatchWriter, StreamWriter, Filter, MultiFilter
from idc.api import make_list


class Tee(Filter):
    """
    Forwards the data coming through to the sub-flow.
    """

    def __init__(self, sub_flow: List[Plugin] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param sub_flow: the filter(s)/writer to forward the data to
        :type sub_flow: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.sub_flow = sub_flow
        self._filter = None
        self._writer = None
        self._data_buffer = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "tee"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Forwards the data passing through to the filter/writer defined as its sub-flow."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--sub_flow", type=str, default=None, help="The command-line defining the subflow (filter(s)/writer).")
        return parser

    def _parse_commandline(self, cmdline: str) -> List[Plugin]:
        """
        Parses the command-line and returns the list of plugins it represents.
        Raises an exception in case of an invalid sub-flow.
        
        :param cmdline: the command-line to parse
        :type cmdline: str 
        :return: 
        """
        from idc.registry import available_filters, available_writers
        from seppl import args_to_objects

        # split command-line into valid plugin subsets
        valid = dict()
        valid.update(available_filters())
        valid.update(available_writers())
        args = split_args(split_cmdline(cmdline), list(valid.keys()))
        return args_to_objects(args, valid, allow_global_options=False)

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.sub_flow = None
        if ns.sub_flow is not None:
            self.sub_flow = self._parse_commandline(ns.sub_flow)

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()

        if self.sub_flow is None:
            self.sub_flow = []

        if len(self.sub_flow) > 0:
            self._filter = None
            self._writer = None
            filters = []
            for plugin in self.sub_flow:
                if isinstance(plugin, Filter):
                    filters.append(plugin)
                elif isinstance(plugin, Writer):
                    self._writer = plugin
                    # no more plugins allowed beyond writer
                    break
            if len(filters) == 1:
                self._filter = filters[0]
            elif len(filters) > 1:
                self._filter = MultiFilter(filters=filters)
            if (self._writer is not None) and isinstance(self._writer, BatchWriter):
                self._data_buffer = []

        if self._filter is not None:
            self._filter.session = self.session
            self._filter.initialize()
        if self._writer is not None:
            self._writer.session = self.session
            self._writer.initialize()

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []
        for item in make_list(data):
            # filter data
            if self._filter is not None:
                item = self._filter.process(item)

            # write data
            if (item is not None) and (self._writer is not None):
                if isinstance(self._writer, BatchWriter):
                    self._data_buffer.append(item)
                elif isinstance(self._writer, StreamWriter):
                    self._writer.write_stream(item)
                else:
                    raise Exception("Unhandled type of writer: %s" % str(type(self._writer)))

            result.append(item)

        return result

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()

        # flush batch buffer
        if (self._data_buffer is not None) and isinstance(self._writer, BatchWriter):
            self.logger().info("Flushing data buffer...")
            self._writer.write_batch(self._data_buffer)
            self._data_buffer = None

        # finalize sub-flow
        if self._filter is not None:
            self._filter.finalize()
        if (self._writer is not None) and isinstance(self._writer, Initializable):
            self._writer.finalize()



import argparse
import copy
from typing import List, Union

from seppl.io import Filter
from wai.common.adams.imaging.locateobjects import LocatedObjects, NormalizedLocatedObjects
from wai.logging import LOGGING_WARNING

from idc.api import ObjectDetectionData, ImageClassificationData, ImageSegmentationData, get_object_label, \
    set_object_label, ImageSegmentationAnnotations, flatten_list, make_list


class MapLabels(Filter):
    """
    Filters out labels according to the parameters.
    """

    def __init__(self, mapping: List[str] = None, 
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param mapping: the label mapping to use (list of old=new pairs)
        :type mapping: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.mapping = mapping
        self._mapping = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "map-labels"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Maps labels from one set to another."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData, ImageClassificationData, ImageSegmentationData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData, ImageClassificationData, ImageSegmentationData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-m", "--mapping", type=str, metavar="old=new", default=None, help="The labels to use", required=False, nargs="*")
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.mapping = ns.mapping

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._mapping = {}
        for map_string in self.mapping:
            old, new = map_string.split("=")
            # Make sure we don't double-map a label
            if old in self._mapping:
                raise ValueError(f"Multiple mappings specified for label '%s': %s, %s" % (old, self._mapping[old], new))
            self._mapping[old] = new

    def process_object_detection(self, annotation: Union[LocatedObjects, NormalizedLocatedObjects]) -> bool:
        """
        Maps the labels in the located objects from their current value to
        their new value.

        :param annotation: The parsed objects
        :return: whether updated
        :rtype: bool
        """
        # Do nothing if no mapping provided
        if len(self._mapping) == 0:
            return False

        # Process each object
        result = False
        for located_object in annotation:
            # Get the object's current label
            label = get_object_label(located_object, default_label=None)

            # If the object doesn't have a label, skip it
            if label is None:
                continue

            # If there is a mapping for this label, change it
            if label in self._mapping:
                set_object_label(located_object, self._mapping[label])
                result = True

        return result

    def process_image_segmentation(self, annotation: ImageSegmentationAnnotations) -> bool:
        """
        Processes the image segmentation annotations.

        :param annotation: the annotations to process
        :type annotation: ImageSegmentationAnnotations
        :return: whether the annotations were updated
        :rtype: bool
        """
        result = False

        for label in self._mapping:
            if label in annotation.labels:
                i = annotation.labels.index(label)
                annotation.labels[i] = self._mapping[label]
                result = True
            if label in annotation.layers:
                data = annotation.layers[label]
                del annotation.layers[label]
                annotation.layers[self._mapping[label]] = data
                result = True

        return result

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            item_new = item
            if item.has_annotation():
                if isinstance(item, ObjectDetectionData):
                    ann_new = copy.deepcopy(item.annotation)
                    updated = self.process_object_detection(ann_new)
                    if updated:
                        item_new = item.duplicate(annotation=ann_new)
                elif isinstance(item, ImageClassificationData):
                    if item.annotation in self._mapping:
                        item_new = item.duplicate(annotation=self._mapping[item.annotation])
                elif isinstance(item, ImageSegmentationData):
                    ann_new = copy.deepcopy(item.annotation)
                    updated = self.process_image_segmentation(ann_new)
                    if updated:
                        item_new = item.duplicate(annotation=ann_new)
                else:
                    self.logger().warning("Cannot process data type: %s" % str(type(item)))
            result.append(item_new)

        return flatten_list(result)



from seppl.io import Filter


def parse_filter(filter_: str) -> Filter:
    """
    Parses the command-line and instantiates the filter.

    :param filter_: the command-line to parse
    :type filter_: str
    :return: the filter
    :rtype: Filter
    """
    from seppl import split_args, split_cmdline, args_to_objects
    from idc.registry import available_filters

    if filter_ is None:
        raise Exception("No filter command-line supplied!")
    valid = dict()
    valid.update(available_filters())
    args = split_args(split_cmdline(filter_), list(valid.keys()))
    objs = args_to_objects(args, valid, allow_global_options=False)
    if len(objs) == 1:
        if isinstance(objs[0], Filter):
            result = objs[0]
        else:
            raise Exception("Expected instance of Filter but got: %s" % str(type(objs[0])))
    else:
        raise Exception("Expected to obtain one filter from '%s' but got %d instead!" % (filter_, len(objs)))
    return result



import argparse
import re
from typing import List

from seppl.io import Filter
from shapely.geometry import Polygon
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from wai.logging import LOGGING_WARNING

from idc.api import ObjectDetectionData, locatedobject_polygon_to_shapely, intersect_over_union, get_object_label, flatten_list, make_list


class LabelPresent(Filter):
    """
    Only forwards images that have the specified label(s) present.
    """

    def __init__(self, labels: List[str] = None, regexp: str = None, region: str = None,
                 coordinate_separator: str = ";", pair_separator: str = ",", min_iou: float = None,
                 invert_regions: bool = False, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the filter.

        :param labels: the labels to use
        :type labels: list
        :param regexp: the regular expression for using only a subset
        :type regexp: str
        :param region: list of x/y pairs defining the region that the object must overlap with in order to be included. Values between 0-1 are considered normalized, otherwise absolute pixels.
        :type region: str
        :param coordinate_separator: the separator between coordinates
        :type coordinate_separator: str
        :param pair_separator: the separator between coordinates
        :type pair_separator: str
        :param min_iou: the minimum IoU (intersect over union) that the object must have with the region in order to be considered an overlap (object detection only)
        :type min_iou: float
        :param invert_regions: Inverts the matching sense from 'labels have to overlap at least one of the region(s)' to 'labels cannot overlap any region'
        :type invert_regions: bool
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.labels = labels
        self.regexp = regexp
        self.region = region
        self.coordinate_separator = coordinate_separator
        self.pair_separator = pair_separator
        self.min_iou = min_iou
        self.invert_regions = invert_regions
        self._pattern = None
        self._regions = None
        self._normalized = None
        self._polygons = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "label-present"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Only forwards images that have the specified label(s) present."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("--labels", type=str, default=None, help="The labels to use", required=False, nargs="*")
        parser.add_argument("--regexp", type=str, default=None, help="Regular expression for using only a subset of labels", required=False)
        parser.add_argument("--region", type=str, default=None, help="List of x/y pairs defining the region that the object must overlap with in order to be included. Values between 0-1 are considered normalized, otherwise absolute pixels.", required=False)
        parser.add_argument("--coordinate_separator", type=str, default=";", help="the separator between coordinates", required=False)
        parser.add_argument("--pair_separator", type=str, default=",", help="the separator between the x and y of a pair", required=False)
        parser.add_argument("--min_iou", type=float, default=None, help="The minimum IoU (intersect over union) that the object must have with the region in order to be considered an overlap (object detection only)", required=False)
        parser.add_argument("--invert_regions", action="store_true", help="Inverts the matching sense from 'labels have to overlap at least one of the region(s)' to 'labels cannot overlap any region'", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.labels = ns.labels
        self.regexp = ns.regexp
        self.region = ns.region
        self.coordinate_separator = ns.coordinate_separator
        self.pair_separator = ns.pair_separator
        self.min_iou = ns.min_iou
        self.invert_regions = ns.invert_regions

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._pattern = re.compile(self.regexp) if (self.regexp is not None) else None

        if len(self.coordinate_separator) != 1:
            raise Exception("Coordinate separator must be a single character, but found: %s" % self.coordinate_separator)
        if len(self.pair_separator) != 1:
            raise Exception("Pair separator must be a single character, but found: %s" % self.pair_separator)

        self._regions = []
        self._normalized = True
        self._polygons = {}
        if self.region is not None:
            for r in self.region:
                region = []
                coords = r.split(self.coordinate_separator)
                if len(coords) < 3:
                    raise Exception("Region must have at least three coordinates of format 'x,y' separated by '%s', but found: %s" % (self.coordinate_separator, r))
                for coord in coords:
                    pair = coord.split(self.pair_separator)
                    if len(pair) != 2:
                        raise Exception("Coordinates must have format 'x%sy', but found '%s' in region '%s'!" % (self.pair_separator, coord, r))
                    region.append([float(x) for x in pair])
                self._regions.append(region)
                # not normalized?
                if sum([(0 if (sum(x) < 2) else 1) for x in region]) > 0:
                    self._normalized = False

    def check_label(self, label: str) -> bool:
        """
        Checks whether the label fits the criteria.

        :param label:   The label to test.
        :return:        True if the label matches, false if not.
        """
        if len(self.labels) == 0 and self._pattern is None:
            return True
        elif len(self.labels) == 0:
            return bool(self._pattern.match(label))
        elif self._pattern is None:
            return label in self.labels
        else:
            return bool(self._pattern.match(label)) or label in self.labels

    def check_regions(self, located_object: LocatedObject, width: int, height: int) -> bool:
        """
        Checks whether the object falls within at least one of the regions or, when inverting,
        whether it does not fall within any.

        :param located_object: the object to check
        :param width: the width of the image
        :param height: the height of the image
        :return: True if no regions defined or if object matches at least one region (invert_regions=False) or none at all (invert_regions=True)
        """
        if len(self._regions) == 0:
            return True

        key = "%d-%d" % (width, height)

        # already created polygons for image dimensions?
        if key not in self._polygons:
            region_polys = []
            self._polygons[key] = region_polys
            for region in self._regions:
                if self._normalized:
                    points = []
                    for x, y in region:
                        points.append([int(x * width), int(y * height)])
                else:
                    points = region
                region_polys.append(Polygon(points))
        else:
            region_polys = self._polygons[key]

        # overlap with any region?
        object_poly = locatedobject_polygon_to_shapely(located_object)
        match = False
        for region_poly in region_polys:
            iou = intersect_over_union(object_poly, region_poly)
            self.logger().debug("object_poly: %s" % str(object_poly))
            self.logger().debug("region_poly: %s" % str(region_poly))
            self.logger().debug("iou: %f > %f = %s" % (iou, self.min_iou, str(iou > self.min_iou)))
            if iou > self.min_iou:
                match = True
                break

        result = (not self.invert_regions and match) or (self.invert_regions and not match)
        self.logger().debug("check_regions = %s" % str(result))
        return result

    def find_valid_objects(self, located_objects: LocatedObjects, width: int, height: int) -> List[int]:
        """
        Returns indices of objects that match the search criteria.

        :param located_objects: The located objects to process.
        :param width: the width of the image
        :param height: the height of the image
        :return: the list of indices of objects matching the criteria
        """
        result = []

        # Search the located objects
        for index, located_object in enumerate(located_objects):
            label = get_object_label(located_object)
            label_ok = self.check_label(label)
            self.logger().debug("check_label: %s = %s" % (label, str(label_ok)))
            if label_ok:
                if self.check_regions(located_object, width, height):
                    result.append(index)

        return result

    def _do_process(self, data):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []

        for item in make_list(data):
            # determines annotations that match criteria
            indices = self.find_valid_objects(item.get_absolute(), item.image_width, item.image_height)
            self.logger().debug("indices: %s" % str(indices))
            if len(indices) > 0:
                result.append(item)
            else:
                self.logger().info("Discarding: %s" % item.image_name)

        return flatten_list(result)



import argparse
from typing import List

from wai.logging import LOGGING_WARNING

from idc.api import DATATYPES, data_type_to_class, ImageData, SplittableStreamWriter, make_list, load_function


class PythonFunctionWriter(SplittableStreamWriter):

    def __init__(self, function: str = None, data_type: str = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param function: the function to use (module_name:function_name)
        :type function: str
        :param data_type: the type of output to generate from the images
        :type data_type: str
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.function = function
        self.data_type = data_type
        self._input_cls = None
        self._function = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-pyfunc"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Processes the images of the specified data type via the declared Python function. The function must take an image container as input (matching the data type class) and an optional 'split' string parameter."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-f", "--function", type=str, default=None, help="The Python function to use, format: module_name:function_name", required=True)
        parser.add_argument("-t", "--data_type", choices=DATATYPES, type=str, default=None, help="The type of data to forward", required=True)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.function = ns.function
        self.data_type = ns.data_type

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._input_cls = data_type_to_class(self.data_type)
        self._function = load_function(self.function)

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        for item in make_list(data):
            if issubclass(type(item), self._input_cls):
                split = None
                if self.splitter is not None:
                    split = self.splitter.next(item=item.image_name)
                self.logger().info("Processing image '%s' (split: %s)" % (item.image_name, str(split)))
                self._function(item, split)
            else:
                self.logger().error("Did not receive an object of type '%s' but of type '%s'!" % (self._input_cls, type(item)))



from typing import List, Dict


def list_classes() -> Dict[str, List[str]]:
    return {
        "seppl.io.Reader": [
            "idc.reader",
            "idc.reader.imgcls",
            "idc.reader.imgseg",
            "idc.reader.objdet",
        ],
        "seppl.io.Filter": [
            "idc.filter",
            "idc.filter.imgcls",
            "idc.filter.imgseg",
            "idc.filter.objdet",
        ],
        "seppl.io.Writer": [
            "idc.writer",
            "idc.writer.imgcls",
            "idc.writer.imgseg",
            "idc.writer.objdet",
        ],
        "idc.api.Generator": [
            "idc.generator",
        ],
    }



import argparse
import json
import os
from datetime import datetime
from typing import List, Iterable, Dict

from wai.logging import LOGGING_WARNING

from idc.api import ObjectDetectionData, SplittableBatchWriter, get_object_label, AnnotationsOnlyWriter, add_annotations_only_param
from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter


class COCOObjectDetectionWriter(SplittableBatchWriter, AnnotationsOnlyWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_dir: str = None,
                 license_name: str = "default", license_url: str = "",
                 categories: List[str] = None, error_on_new_category: bool = False,
                 default_supercategory: str = "Object", sort_categories: bool = False,
                 category_output_file: str = None, annotations_only: bool = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the writer.

        :param output_dir: the output directory to save the image/report in
        :type output_dir: str
        :param license_name: the name of the license to use
        :type license_name: str
        :param license_url: the URL of the license to use
        :type license_url: str
        :param categories: the predefined categories to use
        :type categories: str
        :param error_on_new_category: whether to raise an exception when an unseen category is encountered
        :type error_on_new_category: bool
        :param default_supercategory: the default super category to use, eg Object
        :type default_supercategory: str
        :param sort_categories: whether to sort the categories
        :type sort_categories: bool
        :param category_output_file: the file to store the categories in (comma-separated list)
        :type category_output_file: str
        :param annotations_only: whether to output only the annotations and not the images
        :type annotations_only: bool
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.license_name = license_name
        self.license_url = license_url
        self.categories = categories
        self.error_on_new_category = error_on_new_category
        self.default_supercategory = default_supercategory
        self.sort_categories = sort_categories
        self.category_output_file = category_output_file
        self.annotations_only = annotations_only
        self._category_lookup = None
        self._image_lookup = None
        self._splits = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-coco-od"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the bounding box/polygon definitions in MS COCO .json format."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the images/.json files in. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        parser.add_argument("--license_name", type=str, help="The name of the license to use.", required=False, default="default")
        parser.add_argument("--license_url", type=str, help="The URL of the license to use.", required=False, default="")
        parser.add_argument("--categories", type=str, help="The predefined order of categories.", required=False, nargs="*")
        parser.add_argument("--error_on_new_category", action="store_true", help="Whether to raise an exception if an unknown category is encountered.", required=False)
        parser.add_argument("--default_supercategory", type=str, help="The default super category to use, e.g., 'Object'.", required=False, default="Object")
        parser.add_argument("--sort_categories", action="store_true", help="Whether to sort the categories.", required=False)
        parser.add_argument("--category_output_file", type=str, help="The name of the file (no path) to store the categories in as comma-separated list.", required=False, default=None)
        add_annotations_only_param(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.license_name = ns.license_name
        self.license_url = ns.license_url
        self.categories = ns.categories
        self.error_on_new_category = ns.error_on_new_category
        self.default_supercategory = ns.default_supercategory
        self.sort_categories = ns.sort_categories
        self.category_output_file = ns.category_output_file
        self.annotations_only = ns.annotations_only

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()

        if self.annotations_only is None:
            self.annotations_only = False

        self._category_lookup = dict()
        if self.categories is not None:
            cats = self.categories
            if self.sort_categories:
                cats = sorted(cats)
            for i, category in enumerate(cats, start=1):
                self._category_lookup[category] = i
        else:
            if self.sort_categories:
                self.logger().warning("Sorting of categories is only supported when using predefined categories!")
                self.sort_categories = False

        self._splits = dict()
        self._image_lookup = dict()

    def _create_info(self) -> Dict:
        """
        Creates the info section.

        :return: the info section
        :rtype: dict
        """
        result = dict()
        result["year"] = datetime.now().strftime("%Y")
        result["version"] = ""
        result["description"] = "Converted with image-dataset-converter"
        result["contributor"] = ""
        result["url"] = ""
        result["date_created"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
        return result

    def _create_licenses(self) -> List:
        """
        Creates the licenses section.

        :return: the licenses section
        :rtype: list
        """
        result = list()
        result.append(dict())
        result[0]["id"] = 1
        result[0]["name"] = self.license_name
        result[0]["url"] = self.license_url
        return result

    def _create_categories(self) -> List:
        """
        Creates the categories section.

        :return: the categories section
        :rtype: list
        """
        result = list()
        for category_name, category_id in self._category_lookup.items():
            category = dict()
            category["id"] = category_id
            category["name"] = category_name
            category["supercategory"] = self.default_supercategory
            result.append(category)
        return result

    def _init_annotations(self) -> Dict:
        """
        Initializes the annotations structure.
        """
        result = dict()
        result["info"] = self._create_info()
        result["licenses"] = self._create_licenses()
        result["images"] = list()
        result["annotations"] = list()
        result["categories"] = list()
        return result

    def _append_image(self, data: Dict, item):
        """
        Appends the image to the images section.

        :param data: the annotations structure to update
        :type data: dict
        """
        image_id = len(data["images"]) + 1
        self._image_lookup[item.image_name] = image_id
        image = dict()
        image["id"] = image_id
        image["width"] = item.image_width
        image["height"] = item.image_height
        image["file_name"] = item.image_name
        image["license"] = 1
        image["flickr_url"] = ""
        image["coco_url"] = ""
        image["date_captured"] = ""
        data["images"].append(image)

    def _append_annotations(self, data: Dict, item):
        """
        Appends the annotations to the annotations section.

        :param data: the annotations structure to update
        :type data: dict
        """
        image_id = self._image_lookup[item.image_name]
        absolute = item.get_absolute()
        for obj in absolute:
            label = get_object_label(obj)
            if label not in self._category_lookup:
                if self.error_on_new_category:
                    raise Exception("Undefined label encountered with image %s: %s" % (item.image_name, label))
                self._category_lookup[label] = len(self._category_lookup) + 1
            category_id = self._category_lookup[label]
            annotation_id = len(data["annotations"]) + 1
            annotation = dict()
            annotation["id"] = annotation_id
            annotation["image_id"] = image_id
            annotation["category_id"] = category_id
            annotation["area"] = float(item.image_width * item.image_height)
            annotation["bbox"] = [obj.x, obj.y, obj.width, obj.height]
            annotation["iscrowd"] = 0
            if obj.has_polygon():
                x_list = obj.get_polygon_x()
                y_list = obj.get_polygon_y()
            else:
                x_list = [obj.x, obj.x + obj.width - 1, obj.x + obj.width - 1, obj.x]
                y_list = [obj.y, obj.y, obj.y + obj.height - 1, obj.y + obj.height - 1]
            segmentation = []
            for x, y in zip(x_list, y_list):
                segmentation.append(x)
                segmentation.append(y)
            annotation["segmentation"] = [segmentation]
            data["annotations"].append(annotation)

    def write_batch(self, data: Iterable):
        """
        Saves the data in one go.

        :param data: the data to write
        :type data: Iterable
        """
        for item in data:
            sub_dir = self.session.expand_placeholders(self.output_dir)
            if self.splitter is not None:
                split = self.splitter.next()
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating dir: %s" % sub_dir)
                os.makedirs(sub_dir)

            # write image
            path = os.path.join(sub_dir, item.image_name)
            if not self.annotations_only:
                self.logger().info("Writing image to: %s" % path)
                item.save_image(path)

            # append annotations
            if sub_dir not in self._splits:
                self._splits[sub_dir] = self._init_annotations()
            self._append_image(self._splits[sub_dir], item)
            if item.has_annotation():
                self._append_annotations(self._splits[sub_dir], item)

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()

        for sub_dir, annotations in self._splits.items():
            # save annotations
            annotations["categories"] = self._create_categories()
            path = os.path.join(sub_dir, "annotations.json")
            self.logger().info("Writing annotations to: %s" % path)
            with open(path, "w") as fp:
                json.dump(annotations, fp)

            # save categories
            if self.category_output_file is not None:
                categories = self._category_lookup.keys()
                if self.sort_categories:
                    categories = sorted(categories)
                path = os.path.join(sub_dir, self.category_output_file)
                self.logger().info("Writing categories to: %s" % path)
                with open(path, "w") as fp:
                    fp.write(",".join(categories))



from setuptools import setup, find_namespace_packages


def _read(f):
    """
    Reads in the content of the file.
    :param f: the file to read
    :type f: str
    :return: the content
    :rtype: str
    """
    return open(f, 'rb').read()


setup(
    name="image_dataset_converter",
    description="Python3 library for converting between various image annotation dataset formats.",
    long_description=(
            _read('DESCRIPTION.rst') + b'\n' +
            _read('CHANGES.rst')).decode('utf-8'),
    url="https://github.com/waikato-datamining/image-dataset-converter",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Image Processing',
    ],
    license='MIT License',
    package_dir={
        '': 'src'
    },
    packages=find_namespace_packages(where='src'),
    install_requires=[
        "setuptools",
        "seppl>=0.2.15",
        "wai_logging",
        "wai.common>=0.0.44",
        "pillow",
        "matplotlib",
        "imagesize",
        "fast_opex>=0.0.2",
        "defusedxml",
        "numpy<2.0.0",
        "shapely",
        "python_image_complete",
        "simple_palette_utils",
    ],
    version="0.0.10",
    author='Peter Reutemann',
    author_email='fracpete@waikato.ac.nz',
    entry_points={
        "console_scripts": [
            "idc-convert=idc.tool.convert:sys_main",
            "idc-exec=idc.tool.exec:sys_main",
            "idc-find=idc.tool.find:sys_main",
            "idc-help=idc.tool.help:sys_main",
            "idc-registry=idc.registry:sys_main",
            "idc-test-generator=idc.tool.test_generator:sys_main",
        ],
        "class_lister": [
            "idc=idc.class_lister",
        ],
    },
)



from typing import List

from seppl import get_class_name, AnyData
from seppl.io import Filter

from idc.api import ImageData, flatten_list, make_list


class StripAnnotations(Filter):

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "strip-annotations"

    def description(self) -> str:
        """
        Returns a description of the handler.

        :return: the description
        :rtype: str
        """
        return "Removes all annotations from the data coming through."

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [AnyData]

    def _do_process(self, data: ImageData):
        """
        Processes the data record(s).

        :param data: the record(s) to process
        :return: the potentially updated record(s)
        """
        result = []
        for item in make_list(data):
            if isinstance(item, ImageData):
                data = data.duplicate()
                data.annotation = None
                result.append(data)
            else:
                self.logger().warning("Expected %s but got %s instead, failed to strip annotations!"
                                      % (get_class_name(ImageData), get_class_name(data)))

        return flatten_list(result)



import argparse
import csv
from typing import List, Iterable, Union

from wai.logging import LOGGING_WARNING
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from wai.common.geometry import Polygon, Point
from seppl.placeholders import PlaceholderSupporter, placeholder_list
from seppl.io import locate_files
from idc.api import ObjectDetectionData, locate_image
from idc.api import Reader


class ROIObjectDetectionReader(Reader, PlaceholderSupporter):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 suffix: str = "-rois.csv", resume_from: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param suffix: the suffix of the ROI CSV files (eg -rois.csv)
        :type suffix: str
        :param resume_from: the file to resume from (glob)
        :type resume_from: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.suffix = suffix
        self.resume_from = resume_from
        self._inputs = None
        self._current_input = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-roicsv-od"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the bounding box and/or polygon definitions from the associated ROI .csv file."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the CSV file(s) to read; glob syntax is supported; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the CSV files to use; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("--resume_from", type=str, help="Glob expression matching the file to resume from, e.g., '*/012345-rois.csv'", required=False)
        parser.add_argument("-s", "--suffix", metavar="SUFFIX", type=str, default="-rois.csv", help="The suffix used by the ROI CSV files.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.suffix = ns.suffix
        self.resume_from = ns.resume_from

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        if self.suffix is None:
            raise Exception("No suffix defined!")
        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.csv", resume_from=self.resume_from)

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self.finalize()

        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input
        self.logger().info("Reading from: " + str(self.session.current_input))

        image = locate_image(self.session.current_input, suffix=self.suffix)
        if image is None:
            self.logger().warning("No associated image found: %s" % self._current_input)
            yield None

        with open(self.session.current_input, "r") as fp:
            reader = csv.DictReader(fp)

            annotations = LocatedObjects()
            for row in reader:
                if "x0" in row:
                    x0 = int(float(row["x0"]))
                    y0 = int(float(row["y0"]))
                    x1 = int(float(row["x1"]))
                    y1 = int(float(row["y1"]))
                else:
                    x0 = int(float(row["x"]))
                    y0 = int(float(row["y"]))
                    w = int(float(row["w"]))
                    h = int(float(row["h"]))
                    x1 = x0 + w - 1
                    y1 = y0 + h - 1
                meta = dict()
                fields = ["score", "label", "label_str", "minrect_w", "minrect_h"]
                for field in fields:
                    if field in row:
                        value = row[field]
                        if len(value) > 0:
                            if field in ["score", "minrect_w", "minrect_h"]:
                                value = float(value)
                            elif field in ["label"]:
                                value = int(value)
                            meta[field] = value
                if "label_str" in row:
                    meta["type"] = row["label_str"]
                poly = None
                if ("poly_x" in row) and ("poly_y" in row):
                    points = []
                    px_list = row["poly_x"].split(",")
                    py_list = row["poly_y"].split(",")
                    for x, y in zip(px_list, py_list):
                        points.append(Point(int(float(x)), int(float(y))))
                    poly = Polygon(*points)

                obj = LocatedObject(x=x0, y=y0, width=x1-x0+1, height=y1-y0+1, meta=meta)
                if poly is not None:
                    obj.set_polygon(poly)
                annotations.append(obj)

            if len(annotations) == 0:
                annotations = None

        self._current_input = None

        yield ObjectDetectionData(source=image, annotation=annotations)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0



from ._adams import AdamsObjectDetectionWriter
from ._coco import COCOObjectDetectionWriter
from ._opex import OPEXObjectDetectionWriter
from ._roi import ROIObjectDetectionWriter
from ._voc import VOCObjectDetectionWriter
from ._yolo import YoloObjectDetectionWriter



import argparse
import json
import os.path
from typing import List, Iterable, Union, Dict

from seppl.placeholders import PlaceholderSupporter, placeholder_list
from seppl.io import locate_files
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from wai.common.geometry import Point, Polygon
from wai.logging import LOGGING_WARNING

from idc.api import ObjectDetectionData
from idc.api import Reader


class COCOObjectDetectionReader(Reader, PlaceholderSupporter):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 resume_from: str = None, logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param resume_from: the file to resume from (glob)
        :type resume_from: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.resume_from = resume_from
        self._inputs = None
        self._current_input = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-coco-od"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the bounding box or polygon definitions from the associated .json file in MS COCO format."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the JSON file(s) to read; glob syntax is supported; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the JSON files to use; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("--resume_from", type=str, help="Glob expression matching the file to resume from, e.g., '*/012345.json'", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.resume_from = ns.resume_from

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ObjectDetectionData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.json", resume_from=self.resume_from)

    def _create_lookup(self, data: Dict, key: str, key_name: str) -> Dict:
        """
        Generates a lookup from the data.

        :param data: the data to use
        :type data: dict
        :param key: the name of the key to generate the lookup for
        :type key: str
        :return: the lookup (id -> name)
        :rtype: dict
        """
        if key not in data:
            raise Exception("No '%s' available!" % key)
        result = dict()

        for item in data[key]:
            result[item["id"]] = item[key_name]

        return result

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input
        self.logger().info("Reading from: " + str(self.session.current_input))

        # load annotations
        with open(self.session.current_input, "r") as fp:
            data = json.load(fp)

        categories = self._create_lookup(data, "categories", "name")
        license_names = self._create_lookup(data, "licenses", "name")
        license_urls = self._create_lookup(data, "licenses", "url")
        images = self._create_lookup(data, "images", "file_name")
        image_licenses = self._create_lookup(data, "images", "license")

        for image_id in images:
            filename = images[image_id]
            license_name = None
            license_url = None
            if image_id in image_licenses:
                license_id = image_licenses[image_id]
                if license_id in license_names:
                    license_name = license_names[license_id]
                if len(license_name) == 0:
                    license_name = None
                if license_id in license_urls:
                    license_url = license_urls[license_id]
                    if len(license_url) == 0:
                        license_url = None
            img = os.path.join(os.path.dirname(self.session.current_input), filename)
            if not os.path.exists(img):
                self.logger().error("Image file not found for ID #%d: %s" % (image_id, img))
                continue

            lobjs = LocatedObjects()
            file_meta = dict()
            file_meta["image_id"] = image_id
            file_meta["file"] = self.session.current_input
            if license_name is not None:
                file_meta["license_name"] = license_name
            if license_url is not None:
                file_meta["license_url"] = license_url
            for annotation in data["annotations"]:
                if image_id != annotation["image_id"]:
                    continue
                category_id = annotation["category_id"]
                meta = dict()
                meta["type"] = categories[category_id]
                x, y, w, h = annotation["bbox"]
                lobj = LocatedObject(x, y, w, h, **meta)
                lobjs.append(lobj)
                if annotation["iscrowd"] == 0:
                    segmentation = annotation["segmentation"]
                    if isinstance(segmentation, list) and (len(segmentation) > 0):
                        if len(segmentation) > 1:
                            self.logger().warning("More than one polygon defined for annotation id #%d, only using first!" % annotation["id"])
                        for subsegmentation in segmentation:
                            points = []
                            i = 0
                            while i < len(subsegmentation) - 1:
                                points.append(Point(int(subsegmentation[i]), int(subsegmentation[i+1])))
                                i += 2
                            polygon = Polygon(*points)
                            lobj.set_polygon(polygon)
                            # we only process one
                            break

            yield ObjectDetectionData(source=str(img), annotation=lobjs, metadata=file_meta)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0
